@@ -1,7 +1,7 @@
 import logging
 import setproctitle
 
-from shakenfist import config
+from shakenfist.config import config
 from shakenfist import logutil
 
 
@@ -27,8 +27,7 @@ def set_log_level(log, name):
     process_name(name)
 
     # Check for configuration override
-    label = 'LOGLEVEL_' + name.upper()
-    level = config.parsed.get(label)
+    level = config.get('LOGLEVEL_' + name.upper())
     if level:
         numeric_level = getattr(logging, level.upper(), None)
         if not isinstance(numeric_level, int):
