@@ -10,6 +10,7 @@ from shakenfist.daemons import daemon
 from shakenfist import db
 from shakenfist import exceptions
 from shakenfist.images import Image
+from shakenfist.ipmanager import IPManager
 from shakenfist import logutil
 from shakenfist import net
 from shakenfist import scheduler
@@ -17,11 +18,11 @@ from shakenfist import util
 from shakenfist import virt
 from shakenfist.tasks import (QueueTask,
                               DeleteInstanceTask,
-                              ErrorInstanceTask,
                               FetchImageTask,
                               InstanceTask,
                               PreflightInstanceTask,
                               StartInstanceTask,
+                              DefloatNetworkInterfaceTask
                               )
 
 
@@ -29,13 +30,13 @@ LOG, _ = logutil.setup(__name__)
 
 
 def handle(jobname, workitem):
-    log = LOG.withField('workitem', jobname)
+    log = LOG.with_field('workitem', jobname)
     log.info('Processing workitem')
 
     setproctitle.setproctitle(
         '%s-%s' % (daemon.process_name('queues'), jobname))
 
-    instance_uuid = None
+    instance = None
     task = None
     try:
         for task in workitem.get('tasks', []):
@@ -43,16 +44,21 @@ def handle(jobname, workitem):
                 raise exceptions.UnknownTaskException(
                     'Task was not decoded: %s' % task)
 
-            if (InstanceTask.__subclasscheck__(type(task)) or
-                    isinstance(task, FetchImageTask)):
-                instance_uuid = task.instance_uuid()
+            if InstanceTask.__subclasscheck__(type(task)):
+                instance = virt.Instance.from_db(task.instance_uuid())
+                if not instance:
+                    raise exceptions.InstanceNotInDBException(
+                        task.instance_uuid())
 
-            if instance_uuid:
-                log_i = log.withInstance(instance_uuid)
+            if isinstance(task, FetchImageTask):
+                instance = virt.Instance.from_db(task.instance_uuid())
+
+            if instance:
+                log_i = log.with_instance(instance)
             else:
                 log_i = log
 
-            log_i.withField('task_name', task.name()).info('Starting task')
+            log_i.with_field('task_name', task.name()).info('Starting task')
 
             # TODO(andy) Should network events also come through here eventually?
             # Then this can be generalised to record events on networks/instances
@@ -61,128 +67,113 @@ def handle(jobname, workitem):
             # dequeued in the DB. Currently it's reporting action on the item
             # and calling it 'dequeue'.
 
-            if instance_uuid:
+            if instance:
                 # TODO(andy) move to QueueTask
-                db.add_event('instance', instance_uuid, task.pretty_task_name(),
+                db.add_event('instance', instance.uuid, task.pretty_task_name(),
                              'dequeued', None, 'Work item %s' % jobname)
 
             if isinstance(task, FetchImageTask):
-                image_fetch(task.url(), instance_uuid)
+                image_fetch(task.url(), instance)
 
             elif isinstance(task, PreflightInstanceTask):
-                redirect_to = instance_preflight(instance_uuid, task.network())
+                redirect_to = instance_preflight(instance, task.network())
                 if redirect_to:
                     log_i.info('Redirecting instance start to %s'
                                % redirect_to)
-                    db.place_instance(instance_uuid, redirect_to)
                     db.enqueue(redirect_to, workitem)
                     return
 
             elif isinstance(task, StartInstanceTask):
-                instance_start(instance_uuid, task.network())
-                db.update_instance_state(instance_uuid, 'created')
+                instance_start(instance, task.network())
                 db.enqueue('%s-metrics' % config.NODE_NAME, {})
 
             elif isinstance(task, DeleteInstanceTask):
                 try:
-                    instance_delete(instance_uuid)
-                    db.update_instance_state(instance_uuid, 'deleted')
-                except Exception as e:
-                    util.ignore_exception(daemon.process_name('queues'), e)
-
-            elif isinstance(task, ErrorInstanceTask):
-                try:
-                    instance_delete(instance_uuid)
-                    db.update_instance_state(instance_uuid, 'error')
-
-                    if task.error_msg():
-                        db.update_instance_error_message(
-                            instance_uuid, task.error_msg())
+                    instance_delete(instance)
                     db.enqueue('%s-metrics' % config.NODE_NAME, {})
                 except Exception as e:
                     util.ignore_exception(daemon.process_name('queues'), e)
 
             else:
-                log_i.withField('task', task).error('Unhandled task - dropped')
+                log_i.with_field('task', task).error(
+                    'Unhandled task - dropped')
 
             log_i.info('Task complete')
 
     except exceptions.ImageFetchTaskFailedException as e:
         # Usually caused by external issue and not an application error
         log.info('Fetch Image Error: %s', e)
-        if instance_uuid:
-            db.enqueue_instance_error(instance_uuid,
-                                      'failed queue task: %s' % e)
+        if instance:
+            instance.enqueue_delete_due_error('Failed queue task: %s' % e)
 
     except Exception as e:
+        # Logging ignored exception - this should be investigated
         util.ignore_exception(daemon.process_name('queues'), e)
-        if instance_uuid:
-            db.enqueue_instance_error(instance_uuid,
-                                      'failed queue task: %s' % e)
+        if instance:
+            instance.enqueue_delete_due_error('Failed queue task: %s' % e)
 
     finally:
         db.resolve(config.NODE_NAME, jobname)
-        if instance_uuid:
-            db.add_event('instance', instance_uuid, 'tasks complete',
-                         'dequeued', None, 'Work item %s' % jobname)
+        if instance:
+            instance.add_event('tasks complete', 'dequeued',
+                               msg='Work item %s' % jobname)
         log.info('Completed workitem')
 
 
-def image_fetch(url, instance_uuid):
-    instance = None
-    if instance_uuid:
-        instance = virt.from_db(instance_uuid)
-
+def image_fetch(url, instance):
     try:
         # TODO(andy): Wait up to 15 mins for another queue process to download
         # the required image. This will be changed to queue on a
         # "waiting_image_fetch" queue but this works now.
         with db.get_lock('image', config.NODE_NAME, Image.calc_unique_ref(url),
                          timeout=15*60, op='Image fetch') as lock:
-            img = Image.from_url(url)
+            # Note that the image might already exist in the database as the API
+            # creates a records so that the image is included in listings before
+            # it is fetched. However, the new() call here handles that case and
+            # will just return the previous entry if one exists.
+            img = Image.new(url)
             img.get([lock], instance)
             db.add_event('image', url, 'fetch', None, None, 'success')
 
     except (exceptions.HTTPError, requests.exceptions.RequestException) as e:
-        LOG.withField('image', url).info('Failed to fetch image')
-        if instance_uuid:
-            db.enqueue_instance_error(instance_uuid,
-                                      'Image fetch failed: %s' % e)
+        LOG.with_field('image', url).info('Failed to fetch image')
 
         # Clean common problems to store in events
         msg = str(e)
-        re_conn_err = re.compile(r'.*NewConnectionError\(\'\<.*\>: (.*)\'')
-        m = re_conn_err.match(msg)
-        if m:
-            msg = m.group(1)
-        db.add_event('image', url, 'fetch', None, None, 'Error: '+msg)
+        if msg.find('Name or service not known'):
+            msg = 'DNS error'
+        if msg.find('No address associated with hostname'):
+            msg = 'DNS error'
+        db.add_event('image', url, 'fetch', None, None, msg)
 
         raise exceptions.ImageFetchTaskFailedException(
-            'Failed to fetch image %s' % url)
+            'Failed to fetch image: %s Exception: %s' % (url, e))
 
 
-def instance_preflight(instance_uuid, network):
-    db.update_instance_state(instance_uuid, 'preflight')
+def instance_preflight(instance, network):
+    instance.state = 'preflight'
 
+    # Try to place on this node
     s = scheduler.Scheduler()
-    instance = virt.from_db(instance_uuid)
-
     try:
         s.place_instance(instance, network, candidates=[config.NODE_NAME])
         return None
 
     except exceptions.LowResourceException as e:
-        db.add_event('instance', instance_uuid,
-                     'schedule', 'retry', None,
-                     'insufficient resources: ' + str(e))
+        instance.add_event('schedule', 'retry', None,
+                           'insufficient resources: ' + str(e))
 
-    if instance.db_entry.get('placement_attempts') > 3:
+    # Unsuccessful placement, check if reached placement attempt limit
+    db_placement = instance.placement
+    if db_placement['placement_attempts'] > 3:
         raise exceptions.AbortInstanceStartException(
             'Too many start attempts')
 
+    # Try placing on another node
     try:
-        if instance.db_entry.get('requested_placement'):
-            candidates = [instance.db_entry.get('requested_placement')]
+        if instance.requested_placement:
+            # TODO(andy): Ask Mikal why this is not the current node?
+            candidates = [instance.requested_placement]
         else:
             candidates = []
             for node in s.metrics.keys():
@@ -191,51 +182,38 @@ def instance_preflight(instance_uuid, network):
 
         candidates = s.place_instance(instance, network,
                                       candidates=candidates)
+        instance.place_instance(candidates[0])
         return candidates[0]
 
     except exceptions.LowResourceException as e:
-        db.add_event('instance', instance_uuid,
-                     'schedule', 'failed', None,
-                     'insufficient resources: ' + str(e))
+        instance.add_event('schedule', 'failed', None,
+                           'insufficient resources: ' + str(e))
         # This raise implies delete above
         raise exceptions.AbortInstanceStartException(
             'Unable to find suitable node')
 
 
-def instance_start(instance_uuid, network):
-    log = LOG.withField('instance', instance_uuid)
-
-    with db.get_lock(
-            'instance', None, instance_uuid, ttl=900, timeout=120,
-            op='Instance start') as lock:
-        instance = virt.from_db(instance_uuid)
-
-        # Collect the networks
+def instance_start(instance, network):
+    with instance.get_lock(ttl=900, op='Instance start') as lock:
+        # Ensure networks are connected to this node
         nets = {}
         for netdesc in network:
             if netdesc['network_uuid'] not in nets:
-                n = net.from_db(netdesc['network_uuid'])
+                n = net.Network.from_db(netdesc['network_uuid'])
                 if not n:
-                    db.enqueue_instance_error(instance_uuid, 'missing network')
+                    instance.enqueue_delete_due_error(
+                        'missing network: %s' % netdesc['network_uuid'])
                     return
 
-                nets[netdesc['network_uuid']] = n
-
-        # Create the networks
-        with util.RecordedOperation('ensure networks exist', instance):
-            for network_uuid in nets:
-                n = nets[network_uuid]
-                try:
-                    n.create()
-                    n.ensure_mesh()
-                    n.update_dhcp()
-                except exceptions.DeadNetwork as e:
-                    log.withField('network', n).warning(
-                        'Instance tried to use dead network')
-                    db.enqueue_instance_error(
-                        instance_uuid, 'tried to use dead network: %s' % e)
+                if n.state.value != 'created':
+                    instance.enqueue_delete_due_error(
+                        'network is not active: %s' % n.uuid)
                     return
 
+                n.create_on_hypervisor()
+                n.ensure_mesh()
+                n.update_dhcp()
+
         # Allocate console and VDI ports
         instance.allocate_instance_ports()
 
@@ -250,51 +228,65 @@ def instance_start(instance_uuid, network):
             code = e.get_error_code()
             if code in (libvirt.VIR_ERR_CONFIG_UNSUPPORTED,
                         libvirt.VIR_ERR_XML_ERROR):
-                db.enqueue_instance_error(instance_uuid,
-                                          'instance failed to start: %s' % e)
+                instance.enqueue_delete_due_error(
+                    'instance failed to start: %s' % e)
                 return
 
-        for iface in db.get_instance_interfaces(instance_uuid):
+        for iface in db.get_instance_interfaces(instance.uuid):
             db.update_network_interface_state(iface['uuid'], 'created')
 
 
-def instance_delete(instance_uuid):
-    with db.get_lock('instance', None, instance_uuid, timeout=120,
-                     op='Instance delete'):
-        db.add_event('instance', instance_uuid,
-                     'queued', 'delete', None, None)
+def instance_delete(instance):
+    with instance.get_lock(op='Instance delete'):
+        db.add_event('instance', instance.uuid, 'queued', 'delete', None, None)
 
         # Create list of networks used by instance
         instance_networks = []
-        for iface in list(db.get_instance_interfaces(instance_uuid)):
+        for iface in list(db.get_instance_interfaces(instance.uuid)):
             if not iface['network_uuid'] in instance_networks:
                 instance_networks.append(iface['network_uuid'])
 
         # Create list of networks used by all other instances
         host_networks = []
-        for inst in list(db.get_instances(only_node=config.NODE_NAME)):
-            if not inst['uuid'] == instance_uuid:
-                for iface in db.get_instance_interfaces(inst['uuid']):
+        for inst in virt.Instances([virt.this_node_filter,
+                                    virt.active_states_filter]):
+            if not inst.uuid == instance.uuid:
+                for iface in db.get_instance_interfaces(inst.uuid):
                     if not iface['network_uuid'] in host_networks:
                         host_networks.append(iface['network_uuid'])
 
-        instance_from_db_virt = virt.from_db(instance_uuid)
-        if instance_from_db_virt:
-            instance_from_db_virt.delete()
+        instance.delete()
+
+        # Delete the instance's interfaces
+        with util.RecordedOperation('release network addresses', instance):
+            for ni in db.get_instance_interfaces(instance.uuid):
+                if ni.get('floating'):
+                    db.enqueue(
+                        'networknode',
+                        DefloatNetworkInterfaceTask(ni['network_uuid'], ni['uuid']))
+
+                with db.get_lock('ipmanager', None, ni['network_uuid'],
+                                 ttl=120, op='Instance delete'):
+                    ipm = IPManager.from_db(ni['network_uuid'])
+                    ipm.release(ni['ipv4'])
+                    ipm.persist()
+
+                db.update_network_interface_state(ni['uuid'], 'deleted')
 
         # Check each network used by the deleted instance
         for network in instance_networks:
-            n = net.from_db(network)
+            n = net.Network.from_db(network)
             if n:
                 # If network used by another instance, only update
                 if network in host_networks:
                     with util.RecordedOperation('deallocate ip address',
-                                                instance_from_db_virt):
+                                                instance):
                         n.update_dhcp()
                 else:
                     # Network not used by any other instance therefore delete
-                    with util.RecordedOperation('remove network', n):
-                        n.delete()
+                    with util.RecordedOperation('remove network from node', n):
+                        n.delete_on_node_with_lock()
+        return instance
 
 
 class Monitor(daemon.Daemon):
@@ -303,7 +295,7 @@ class Monitor(daemon.Daemon):
         LOG.info('Starting Queues')
 
         libvirt = util.get_libvirt()
-        conn = libvirt.open(None)
+        conn = libvirt.open('qemu:///system')
         present_cpus, _, _ = conn.getCPUMap()
 
         while True:
