@@ -6,6 +6,29 @@ class VersionSpecificationError(Exception):
     pass
 
 
+# Objects
+class ObjectException(Exception):
+    pass
+
+
+class InvalidStateException(ObjectException):
+    pass
+
+# Instance
+
+
+class InstanceException(Exception):
+    pass
+
+
+class InstanceNotInDBException(InstanceException):
+    pass
+
+
+class InstanceBadDiskSpecification(InstanceException):
+    pass
+
+
 class SchedulerException(Exception):
     pass
 
@@ -39,7 +62,7 @@ class ReadException(DatabaseException):
     pass
 
 
-class BadMetadataPacket(DatabaseException):
+class BadObjectVersion(DatabaseException):
     pass
 
 
@@ -85,6 +108,10 @@ class NoNetworkTaskException(TaskException):
     pass
 
 
+class NoNetworkInterfaceTaskException(TaskException):
+    pass
+
+
 class NetworkNotListTaskException(TaskException):
     pass
 
@@ -96,3 +123,7 @@ class NetworkException(Exception):
 
 class DeadNetwork(NetworkException):
     pass
+
+
+class CongestedNetwork(NetworkException):
+    pass
